/*
 * obsmcs  --  OBS Master Clock Syncronization
 *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 *
 * Description : Based on GPS Pulse-per-second (PPS) and NMEA formatted
 *               serial input (GPRMC sentence), the program issues OBS
 *               Master Clock synchronizing pulses. The pattern is as
 *               follows:
 *
 *
 *
 *                                  _____       _____
 *               |                 |     |     |     |
 *               |_________________|     |_____|     |_____
 *               0     5    10    15    20    25    30   [second of minute]
 *
 *
 *
 *
 *               GPS PPS pulse assumed to be active high, of TTL level,
 *               and of min 1 ms duration. Leading edge marks second
 *               boundary.
 *
 *                      ___
 *                     |   |
 *               ______|   |_____
 *
 *                     ^
 *                     |___ Second boundary
 *
 *
 *
 *               ( Program is tested with Garmin GPS mod 35-HVS. )
 *
 *
 * PC ports:
 *              LPT1 :    Pin 2 (D0)    = Output, OBS Master Clock calibration
 *                                        signal.
 *                        Pin 10 (ACK)  = Acknowledge interrupt, PPS from GPS
 *                                        The positive edge of a signal on
 *                                        this pin can be used to generate
 *                                        an interrupt (assumed INT7)
 *                        Pin 18 (GND)  = GND
 *
 *              COM1 :    Pin 2 (RXD)   = NMEA sentences from GPS
 *                        Pin 5 (GND)   = GND
 *
 *
 *
 * NMEA sentence used: GPRMC - Recommended Minimum Specific GPS/TRANSIT Data
 *
 *      $GPRMC,<1>,<2>,<3>,<4>,<5>,<6>,<7>,<8>,<9>,<10>,<11>,<12>*hh<CR><LF>
 *      <1>  UTC time of position fix, hhmmss format
 *      <2>  Statuc, A = Valid position, V = NAV receiver warning
 *      <3>  Latitude, ddmm.mmmm format (leading zeros will be transmitted)
 *      <4>  Latitude hemisphere, N or S
 *      <5>  Longitude, dddmm.mmmm format (leading zeros will be transmitted)
 *      <6>  Longitude hemisphere, E or W
 *      <7>  Speed over ground, 0000.0 to 1851.8 knots (leading zeros will be transmitted)
 *      <8>  Course over ground, 000.0 to 359.9 degrees, true (leading zeros will be transmitted)
 *      <9>  UTC date of position fix, ddmmyy format
 *      <10> Magnetic variation, 000.0 to 180.0 degrees (leading zeros will be transmitted)
 *      <11> Magnetic variation direction, E or W (westerly variation adds to true course)
 *      <12> Mode indicator (only output if NMEA 2.30 active), A = Autonomous,
 *           D = Differential, E = Estimated, N = Data not valid
 *
 */



#ifndef MODULE
#define MODULE
#endif
#define _GNU_SOURCE
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <asm/system.h>
#include <asm/io.h>
#include <rtl.h>
#include <rtl_sync.h>
#include <rtl_sched.h>
#include <rtl_core.h>
#include <rtl_fifo.h>
#include "serp.h"
#include "common.h"


void cleanup_module (void);
unsigned int ack_isr(unsigned int, struct pt_regs*);
unsigned int ser_isr(unsigned int, struct pt_regs*);


static unsigned short seconds=0;
static unsigned int lpt1_data=0;
static unsigned int reset_flag=0;
static unsigned int write_flag=0;


/* 
 *  interrupt handler for parallell port 
 */
unsigned int ack_isr(unsigned int irq, struct pt_regs *regs)
{
  if(reset_flag) {        // reset if reset_flag==1
    seconds = 0;
    reset_flag = 0;
  }
  if(seconds >= 60)       // no more than 60 seconds per minute...
    seconds = 0;          // (below) output 5 second pulses
  if(((seconds % 5) == 0) && !(seconds == 5) && !(seconds==10)) {
    lpt1_data ^= 1;
    outb(lpt1_data, LPT_PORT);
  }
  if(seconds == 0) {
    lpt1_data = 0;
    outb(lpt1_data, LPT_PORT);
  }
  if(write_flag)         // send seconds counter back to monitor
    rtf_put(3, &seconds, sizeof(seconds));

  seconds++;             // increase second counter

  /* acknowledge interrupt and return */
  rtl_hard_enable_irq(LPT_IRQ);
  return(0);
}


/* 
 *  command handler
 *  takes care of commands sent from the monitor
 */
int cmd_handler(unsigned int fifo)
{
  unsigned int cmd;

  /* read command (unsigned integer) from rtf2 */
  rtf_get(2, &cmd, sizeof(int));

  switch(cmd) {
  case 0: // STOP
    break;
  case 1: // RESET
    reset_flag = 1;
    break;
  case 2: // OUTPUT SECONDS ON
    write_flag = 1;
    break;
  case 3: // OUTPUT SECONDS OFF
    write_flag = 0;
    break;
  default: // UNKNOWN COMMAND
    rtl_printf("obsmcs.o: ERR: COMMAND %d UNKNOWN\n", cmd);
    break;
  }
  return 0;
}


////////////////////////////////////////////////////////////////////////////

/* 
 *  initialize resources
 */
int init_module (void)
{
  int i;

  /* create /dev/rtf2  (for commands) */
  i = rtf_create(2, 4096);
  if(i != 0) {
    rtl_printf("obsmcs.o: ERR: can not create fifo 2 (err %d)\n", i);
    return 1;
  }

  /* create /dev/rtf3  (for sending second counter to user space) */
  i = rtf_create(3, 4096);
  if(i != 0) {
    rtl_printf("obsmcs.o: ERR: can not create fifo 3 (err %d)\n", i);
    return 1;
  }

  /* initialize the command handler for rtf2 */
  i = rtf_create_handler(2, &cmd_handler);
  if(i != 0) {
    rtl_printf("obsmcs.o: ERR: can not create handler 2 (err %d)\n", i);
    return 1;
  }

  /* parallel port setup */
  rtl_hard_disable_irq(LPT_IRQ);
  if(rtl_request_irq(LPT_IRQ, ack_isr) != 0) {
    rtl_printf("obsmcs.o: Couldn't create interrupt handler for parallel port\n");
    return 1;
  }
  rtl_hard_enable_irq(LPT_IRQ);  // enable ACK interrupt
  outb(0x10, LPT_PORT+2);

  /* Let the user know the module got inserted properly */
  rtl_printf("obsmcs.o loaded successfully\n");
  return 0;
}


/* 
 *  free all used resources 
 */
void cleanup_module (void)
{
  rtf_destroy(2);  // destroy FIFO's
  rtf_destroy(3);

  /* Free parallel port IRQ (7) */
  outb(0x00, LPT_PORT+2);
  rtl_hard_disable_irq(LPT_IRQ);
  rtl_free_irq(LPT_IRQ);

  /* Say we're done */
  rtl_printf("obsmcs.o unloaded.\n");
}
/////////////////////////////////////////////////////////////////////////////
